/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/macosx/movie.m: Movie playback support for Mac OS X (also
 * used for iOS).
 */

/*
 * Note that, due to bogosity in the AVFoundation framework, SIL on OSX
 * and iOS can only play movies when stored in independent files (in other
 * words, not stored in a package file) if not using the WebM software
 * decoder.  See sys_movie_open() for details.
 */

#define IN_SYSDEP

#import "src/base.h"
#import "src/graphics.h"
#import "src/math.h"
#import "src/memory.h"
#import "src/sound.h"
#import "src/sound/decode.h"
#import "src/sysdep.h"
#import "src/sysdep/macosx/filedata.h"
#ifdef SIL_PLATFORM_IOS
# import "src/sysdep/ios/internal.h"
#else
# import "src/sysdep/macosx/internal.h"
#endif
#import "src/sysdep/opengl/opengl.h"
#import "src/sysdep/opengl/internal.h"  // We rewrite SysTexture fields on draw.
#import "src/sysdep/posix/files.h"  // AVFoundation bogosity workaround, see below.
#import "src/texture.h"

#import <AVFoundation/AVAsset.h>
#import <AVFoundation/AVAssetReader.h>
#import <AVFoundation/AVAssetReaderOutput.h>
#import <AVFoundation/AVAssetTrack.h>
#import <AVFoundation/AVAudioSettings.h>
#import <AVFoundation/AVMediaFormat.h>
#import <CoreAudio/CoreAudioTypes.h>
#import <CoreMedia/CMSampleBuffer.h>
#import <CoreVideo/CVPixelBuffer.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Data for an opened movie. */
struct SysMovieHandle {
    SysFile *fh;                    // File handle to read from
    int width, height;              // Size of the output image
    double framerate;               // Movie framerate
    int audio_rate;                 // Audio playback rate
    int audio_channels;             // Number of audio channels
    float volume;                   // Current volume
    int texture;                    // Texture for video output
    int sound_channel;              // Sound channel for audio output

    OSType pixel_format;            // Pixel format for video decoding
    CMSampleBufferRef audio_buf;    // Current audio buffer being played
    const int16_t *audio_data;      // Pointer to audio_buf's data
    int audio_buflen;               // Number of samples in audio_buf
    int audio_offset;               // Current sample offset within audio_buf
                                    //    (negative during leading silence)

    AVAsset *asset;                 // Asset handle for movie
    AVAssetReader *video_reader;    // Reader for video stream
    AVAssetReaderTrackOutput *video_output;
    AVAssetReader *audio_reader;    // Reader for audio stream
    AVAssetReaderTrackOutput *audio_output;
    NSAutoreleasePool *nspool;      // Autorelease pool for AVFoundation
};

/* Flag: Has the silfiledata:// URL protocol handler been registered? */
static uint8_t protocol_registered = 0;

/*-----------------------------------------------------------------------*/

/* AVFoundation classes are fairly exception-happy, so we define a macro
 * to easily catch and ignore exceptions: */

#define CATCH(...)  do {        \
    @try {__VA_ARGS__;}         \
    @catch (NSException *e) {   \
        DLOG("Exception: %s",   \
             [[NSString stringWithFormat:@"%@", e] UTF8String]); \
    }                           \
} while (0)

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * calculate_framerate:  Calculate the frame rate for the given movie from
 * the difference between timestamps of consecutive frames.
 *
 * [Parameters]
 *     asset: AVAsset instance for movie.
 *     video: AVTrack reference for movie's video track.
 * [Return value]
 *     Frame rate, or zero if the frame rate could not be calculated.
 */
static double calculate_framerate(AVAsset *asset, AVAssetTrack *video);

/**
 * movie_sound_open:  open() implementation for the movie sound decoder.
 *
 * [Return value]
 *     True on success, false on error.
 */
static int movie_sound_open(SoundDecodeHandle *this);

/**
 * movie_sound_reset:  reset() implementation for the movie sound decoder.
 */
static void movie_sound_reset(SoundDecodeHandle *this);

/**
 * movie_sound_get_pcm:  get_pcm() implementation for the movie sound decoder.
 *
 * [Parameters]
 *     pcm_buffer: Buffer into which to store PCM (signed 16-bit) data.
 *     pcm_len: Number of samples to retrieve.
 *     loop_offset_ret: Pointer to variable to receive the number of
 *         samples skipped backward due to looping.
 * [Return value]
 *     Number of samples stored in pcm_buffer.
 */
static int movie_sound_get_pcm(
    SoundDecodeHandle *this, int16_t *pcm_buffer, int pcm_len,
    int *loop_offset_ret);

/**
 * movie_sound_close:  close() implementation for the movie sound decoder.
 */
static void movie_sound_close(SoundDecodeHandle *this);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysMovieHandle *sys_movie_open(SysFile *fh, int64_t offset, int length,
                               UNUSED int direct_render)
{
    SysMovieHandle *movie;

    if (!protocol_registered) {
        if (!osx_filedata_register()) {
            DLOG("Failed to register silfiledata:/// protocol");
            goto error_return;
        }
        protocol_registered = 1;
    }

    /* First allocate and initialize the movie handle structure. */
    movie = mem_alloc(sizeof(*movie), 0, MEM_ALLOC_TEMP | MEM_ALLOC_CLEAR);
    if (UNLIKELY(!movie)) {
        DLOG("alloc(SysMovieHandle) failed");
        goto error_return;
    }
    movie->fh = fh;
    movie->volume = 1.0f;

    /* Create a texture to hold the video image.  We set a dummy size for now;
     * the texture will be rewritten when we get an actual image.  We supply
     * dummy data to preset the correct pixel type for the data the texture
     * will eventually be receiving. */
    const int tex_format = (opengl_has_formats(OPENGL_FORMAT_BGRA)
                            ? TEX_FORMAT_BGRA8888 : TEX_FORMAT_RGBA8888);
    movie->texture = texture_create_with_data(
        1, 1, &(uint32_t[1]){0}, tex_format, 1, 0, 0);
    if (UNLIKELY(!movie->texture)) {
        DLOG("texture_create() failed");
        goto error_free_movie;
    }

    /* Reserve a sound channel for audio output. */
    movie->sound_channel = sound_reserve_channel();
    if (UNLIKELY(!movie->sound_channel)) {
        DLOG("sound_reserve_channel() failed");
        goto error_destroy_texture;
    }

    /* Create an autorelease pool so AVFoundation doesn't leak memory. */
    movie->nspool = [[NSAutoreleasePool alloc] init];
    if (UNLIKELY(!movie->nspool)) {
        DLOG("Failed to create NSAutoreleasePool");
        goto error_free_sound_channel;
    }

    /* Create a AVAsset object for the data stream. */
    NSURL *url = [NSURL alloc];
    if (UNLIKELY(!url)) {
        DLOG("No memory for URL");
        goto error_free_nspool;
    }
    /* In order to read from an arbitrary starting file offset, we use a
     * custom URL protocol with the scheme "silfiledata://" (see filedata.m).
     * However, it seems AVFoundation--like its predecessor QTKit--bypasses
     * the normal URL resolution process entirely, only handling a small,
     * predefined set of URL schemes (file://, http://, and similar standard
     * schemes).  Due to this bogosity, we can only play movies stored in
     * independent files, and for such movies, we need to use a direct
     * file:// URL rather than reading from the file handle passed to this
     * function.  See: http://stackoverflow.com/questions/12963606 */
    if (offset == 0 && length == sys_file_size(fh)) {
        const char *path = posix_file_path(fh);
        ASSERT(path != NULL, goto error_free_nspool);
        NSString *pathstr = [[NSString alloc] initWithUTF8String:path];
        if (UNLIKELY(!pathstr)) {
            DLOG("No memory for path string");
            [url release];
            goto error_free_nspool;
        }
        url = [url initFileURLWithPath:pathstr];
        [pathstr release];
    } else {
        DLOG("Can't play embedded video streams on OSX/iOS"
             " (stackoverflow://12963606)");
        goto error_free_nspool;
        /*
        char urlbuf[100];
        ASSERT(strformat_check(urlbuf, sizeof(urlbuf),
                               "silfiledata:///%p/%lld/%d",
                               fh, (long long)offset, length),
               goto error_free_nspool);
        NSString *urlstr = [[NSString alloc] initWithUTF8String:urlbuf];
        if (UNLIKELY(!urlstr)) {
            DLOG("No memory for URL string");
            [url release];
            goto error_free_nspool;
        }
        url = [url initWithString:urlstr];
        [urlstr release];
        */
    }
    if (UNLIKELY(!url)) {
        DLOG("Failed to initialize URL");
        goto error_free_nspool;
    }
    movie->asset = NULL;
    CATCH(movie->asset = [[AVURLAsset alloc] initWithURL:url options:nil]);
    [url release];
    if (!movie->asset) {
        DLOG("Failed to create asset handle");
        goto error_free_nspool;
    }

    /* Find the video and audio tracks. */
    NSArray *video_tracks =
        [movie->asset tracksWithMediaType:AVMediaTypeVideo];
    if (![video_tracks count]) {
        DLOG("Movie file has no video tracks!");
        goto error_free_asset;
    }
    AVAssetTrack *video = [video_tracks objectAtIndex:0];
    if (UNLIKELY(!video)) {
        DLOG("Couldn't find video track!");
        goto error_free_asset;
    }
    NSArray *audio_tracks =
        [movie->asset tracksWithMediaType:AVMediaTypeAudio];
    AVAssetTrack *audio;
    if (![audio_tracks count]) {
        /* Don't treat this as fatal -- just play with no audio. */
        audio = NULL;
    } else {
        audio = [audio_tracks objectAtIndex:0];
    }

    /* Look up movie attributes. */
    CGSize size = video.naturalSize;
    movie->width = (int)size.width;
    movie->height = (int)size.height;
    if (UNLIKELY(movie->width != size.width || movie->height != size.height)) {
        DLOG("Movie size isn't integral?! (width=%f height=%f)",
             size.width, size.height);
        goto error_free_asset;
    }
    if (UNLIKELY(movie->width == 0 || movie->height == 0)) {
        DLOG("Movie size is zero?! (width=%d height=%d)",
             movie->width, movie->height);
        goto error_free_asset;
    }
    /* Ideally, we could get the frame rate directly from the video track:
     *     movie->framerate = video.nominalFrameRate;
     * Unfortunately, MP4 files generally don't report a correct frame rate,
     * so we calculate it manually by extracting the first two frames and
     * comparing their presentation timestamps. */
    movie->framerate = calculate_framerate(movie->asset, video);
    NSArray *audio_formats;
    int num_audio_formats = 0;
    if (audio) {
        audio_formats = audio.formatDescriptions;
        num_audio_formats = [audio_formats count];
        if (UNLIKELY(!num_audio_formats)) {
            DLOG("Failed to find audio data format!");
            audio = NULL;
        }
    }
    for (int i = 0; i < num_audio_formats; i++) {
        CMAudioFormatDescriptionRef format_desc =
            (CMAudioFormatDescriptionRef)[audio_formats objectAtIndex:i];
        const AudioStreamBasicDescription *audio_desc =
            CMAudioFormatDescriptionGetStreamBasicDescription(format_desc);
        ASSERT(audio_desc != NULL, audio = NULL; break);
        if (!movie->audio_rate) {
            movie->audio_rate = (int)audio_desc->mSampleRate;
            ASSERT(movie->audio_rate > 0, audio = NULL; break);
            movie->audio_channels = (int)audio_desc->mChannelsPerFrame;
            ASSERT(movie->audio_channels > 0, audio = NULL; break);
            if (movie->audio_channels > 2) {
                DLOG("WARNING: %d audio channels not supported",
                     movie->audio_channels);
                audio = NULL;
                break;
            }
        } else if (movie->audio_rate != (int)audio_desc->mSampleRate) {
            DLOG("Warning: found different sample rates for the same audio"
                 " stream (%d vs. %g), using %d", movie->audio_rate,
                 audio_desc->mSampleRate, movie->audio_rate);
            break;  // Avoid spewing multiple errors per track.
        } else if (movie->audio_channels != (int)audio_desc->mChannelsPerFrame) {
            DLOG("Warning: found different channel counts for the same audio"
                 " stream (%d vs. %d), using %d", movie->audio_channels,
                 (int)audio_desc->mChannelsPerFrame, movie->audio_channels);
            break;  // Avoid spewing multiple errors per track.
        }
    }

    /* Create an AVAssetReader for the video stream. */
    NSError *error;
    movie->video_reader = NULL;
    CATCH(movie->video_reader =
        [[AVAssetReader alloc] initWithAsset:movie->asset error:&error]);
    if (UNLIKELY(!movie->video_reader)) {
        DLOG("Failed to create video reader: [%s/%ld] %s",
             [error.domain UTF8String], (long)error.code,
             [error.localizedDescription UTF8String]);
        goto error_free_asset;
    }
    movie->pixel_format = (opengl_has_formats(OPENGL_FORMAT_BGRA)
                           ? kCVPixelFormatType_32BGRA
                           : kCVPixelFormatType_32RGBA);
    NSDictionary *video_settings = @{
        (id)kCVPixelBufferPixelFormatTypeKey: @(movie->pixel_format),
    };
    movie->video_output = NULL;
    CATCH(movie->video_output =
        [[AVAssetReaderTrackOutput alloc] initWithTrack:video
                                          outputSettings:video_settings]);
    if (UNLIKELY(!movie->video_output)) {
        DLOG("Failed to create video reader output");
        goto error_free_video_reader;
    }
    CATCH([movie->video_reader addOutput:movie->video_output]);

    /* Create an AVAssetReader for the audio stream, if any. */
    if (!audio) {
        DLOG("Note: movie has no audio track");
        movie->audio_reader = NULL;
        movie->audio_output = NULL;
    } else {
        movie->audio_reader = NULL;
        CATCH(movie->audio_reader =
            [[AVAssetReader alloc] initWithAsset:movie->asset error:&error]);
        if (UNLIKELY(!movie->audio_reader)) {
            DLOG("Failed to create audio reader: [%s/%ld] %s",
                 [error.domain UTF8String], (long)error.code,
                 [error.localizedDescription UTF8String]);
            goto error_free_video_output;
        }
        NSDictionary *audio_settings = @{
            AVFormatIDKey: @(kAudioFormatLinearPCM),
            AVLinearPCMBitDepthKey: @(16),
#ifdef IS_LITTLE_ENDIAN
            AVLinearPCMIsBigEndianKey: @(NO),
#else
            AVLinearPCMIsBigEndianKey: @(YES),
#endif
            AVLinearPCMIsFloatKey: @(NO),
            AVLinearPCMIsNonInterleavedKey: @(NO),
        };
        movie->audio_output = NULL;
        CATCH(movie->audio_output =
            [[AVAssetReaderTrackOutput alloc] initWithTrack:audio
                                              outputSettings:audio_settings]);
        if (UNLIKELY(!movie->audio_output)) {
            DLOG("Failed to create audio reader output");
            goto error_free_audio_reader;
        }
        CATCH([movie->audio_reader addOutput:movie->audio_output]);
        /* AVFoundation eats some samples from the beginning of the audio
         * track, so fill in an appropriate amount of silence to ensure
         * proper timing. */
#ifdef SIL_PLATFORM_IOS
        if (ios_version_is_at_least("7.0")) {  // Tested through iOS 7.1.
            movie->audio_offset = -3135;
            /* In addition, there's a fade-in of about 1101 samples. */
        } else if (ios_version_is_at_least("5.0")) {
            movie->audio_offset = -5247;
            /* In addition, there's a fade-in of about 1101 samples. */
        } else {
            movie->audio_offset = -1087;
        }
#else  // SIL_PLATFORM_MACOSX
        if (macosx_version_is_at_least(10,9,2)) { // Tested through OSX 10.9.2.
            movie->audio_offset = -9048;
            /* In addition, the first 4608 samples are nulled out. */
        } else if (macosx_version_is_at_least(10,9,1)) {
            movie->audio_offset = -5051;
            /* In addition, the first 4096 samples are nulled out. */
        } else if (macosx_version_is_at_least(10,9,0)) {
            movie->audio_offset = -1880;
            /* In addition, the first 4096 samples are nulled out. */
        } else if (macosx_version_is_at_least(10,8,0)) {
            movie->audio_offset = -5976;
        } else {
            movie->audio_offset = -1087;
        }
#endif
    }

    /* Start the readers running to reduce delay when we actually start
     * playback. */
    [movie->video_reader startReading];
    [movie->audio_reader startReading];

    /* All done. */
    return movie;

  error_free_audio_reader:
    [movie->audio_reader release];
  error_free_video_output:
    [movie->video_output release];
  error_free_video_reader:
    [movie->video_reader release];
  error_free_asset:
    [movie->asset release];
  error_free_nspool:
    [movie->nspool release];
  error_free_sound_channel:
    sound_free_channel(movie->sound_channel);
  error_destroy_texture:
    texture_destroy(movie->texture);
  error_free_movie:
    mem_free(movie);
  error_return:
    sys_file_close(fh);
    return NULL;
}

/*-----------------------------------------------------------------------*/

void sys_movie_close(SysMovieHandle *movie)
{
    sys_movie_stop(movie);
    if (movie->audio_buf) {
        CFRelease(movie->audio_buf);
    }
    [movie->audio_output release];
    [movie->audio_reader release];
    [movie->video_output release];
    [movie->video_reader release];
    [movie->asset release];
    [movie->nspool release];
    sound_free_channel(movie->sound_channel);
    texture_destroy(movie->texture);
    sys_file_close(movie->fh);
    mem_free(movie);
}

/*-----------------------------------------------------------------------*/

int sys_movie_width(SysMovieHandle *movie)
{
    return movie->width;
}

/*-----------------------------------------------------------------------*/

int sys_movie_height(SysMovieHandle *movie)
{
    return movie->height;
}

/*-----------------------------------------------------------------------*/

double sys_movie_framerate(SysMovieHandle *movie)
{
    return movie->framerate;
}

/*-----------------------------------------------------------------------*/

void sys_movie_set_volume(SysMovieHandle *movie, float volume)
{
    movie->volume = volume;
    sound_adjust_volume(movie->sound_channel, volume, 0);
}

/*-----------------------------------------------------------------------*/

int sys_movie_play(SysMovieHandle *movie)
{
    if (movie->audio_reader) {
        SoundDecodeHandle *decoder =
            sound_decode_open_custom(movie_sound_open, movie, 1);
        if (UNLIKELY(!decoder)) {
            DLOG("Failed to create decoder");
            return 0;
        }
        if (UNLIKELY(!sound_play_decoder(decoder, movie->sound_channel,
                                         movie->volume, 0))) {
            DLOG("Failed to start sound");
            sound_decode_close(decoder);
            return 0;
        }
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

void sys_movie_stop(SysMovieHandle *movie)
{
    if (movie->audio_reader) {
        sound_cut(movie->sound_channel);
    }
}

/*-----------------------------------------------------------------------*/

int sys_movie_get_texture(
    SysMovieHandle *movie,
    float *left_ret, float *right_ret, float *top_ret, float *bottom_ret)
{
    *left_ret = 0;
    *right_ret = 1;
    *top_ret = 0;
    *bottom_ret = 1;
    return movie->texture;
}

/*-----------------------------------------------------------------------*/

int sys_movie_draw_frame(SysMovieHandle *movie)
{
    int result = 0;

    if (movie->video_reader.status != AVAssetReaderStatusReading) {
        goto out;
    }

    CMSampleBufferRef sample = 0;
    CATCH(sample = [movie->video_output copyNextSampleBuffer]);
    if (!sample) {
        goto out;
    }
    CVPixelBufferRef image = CMSampleBufferGetImageBuffer(sample);
    if (!image) {
        DLOG("Failed to get video frame image");
        goto release_sample;
    }

    if (UNLIKELY((int)CVPixelBufferGetWidth(image) != movie->width)
     || UNLIKELY((int)CVPixelBufferGetHeight(image) != movie->height)) {
        DLOG("Invalid video image size %dx%d (expected %dx%d)",
             (int)CVPixelBufferGetWidth(image),
             (int)CVPixelBufferGetHeight(image),
             movie->width, movie->height);
        goto release_sample;
    }
    ASSERT(CVPixelBufferGetPixelFormatType(image) == movie->pixel_format,
           goto release_sample);
    ASSERT(!CVPixelBufferIsPlanar(image), goto release_sample);

    CVReturn lock_error = CVPixelBufferLockBaseAddress(image, 0);
    if (UNLIKELY(lock_error != kCVReturnSuccess)) {
        DLOG("Failed to lock video image: %d", (int)lock_error);
        goto release_sample;
    }

    SysTexture *systex = texture_lock_raw(movie->texture);
    if (UNLIKELY(!systex)) {
        DLOG("Failed to lock video texture");
        goto unlock_sample;
    }

    opengl_clear_error();
    glBindTexture(GL_TEXTURE_2D, systex->id);
    const GLenum gl_format = (movie->pixel_format == kCVPixelFormatType_32BGRA
                              ? GL_BGRA : GL_RGBA);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, movie->width, movie->height, 0,
                 gl_format, GL_UNSIGNED_BYTE, CVPixelBufferGetBaseAddress(image));
    if (opengl_current_texture_target == GL_TEXTURE_2D) {
        glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
    } else {
        glBindTexture(GL_TEXTURE_2D, 0);
    }
    const GLenum gl_error = glGetError();
    if (gl_error != GL_NO_ERROR) {
        DLOG("Failed to update video texture: 0x%04X", gl_error);
        goto unlock_sample;
    }

    systex->width = movie->width;
    systex->height = movie->height;
    texture_unlock(movie->texture);
    result = 1;

  unlock_sample:
    ASSERT(CVPixelBufferUnlockBaseAddress(image, 0) == kCVReturnSuccess);
  release_sample:
    CFRelease(sample);
  out:
    return result;
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static double calculate_framerate(AVAsset *asset, AVAssetTrack *video)
{
    NSError *error;
    AVAssetReader *reader =
        [[AVAssetReader alloc] initWithAsset:asset error:&error];
    if (UNLIKELY(!reader)) {
        DLOG("Failed to create video reader: [%s/%ld] %s",
             [error.domain UTF8String], (long)error.code,
             [error.localizedDescription UTF8String]);
        goto error_return;
    }
    AVAssetReaderTrackOutput *output = NULL;
    CATCH(output = [[AVAssetReaderTrackOutput alloc] initWithTrack:video
                                                     outputSettings:nil]);
    if (UNLIKELY(!output)) {
        DLOG("Failed to create video reader output");
        goto error_free_reader;
    }
    CATCH([reader addOutput:output]);

    /* Find the difference between the presentation timestamps for the
     * first two frames.  If the two frames have the same timestamp,
     * continue reading frames until we find one with a different
     * timestamp or we run out of frames. */
    CATCH([reader startReading]);
    CMSampleBufferRef sample1 = 0, sample2 = 0;
    CATCH(sample1 = [output copyNextSampleBuffer]);
    if (UNLIKELY(!sample1)) {
        DLOG("Couldn't retrieve first video frame");
        goto error_free_output;
    }
    const CMTime pts1 = CMSampleBufferGetPresentationTimeStamp(sample1);
    CFRelease(sample1);
    CMTime pts2;
    do {
        sample2 = 0;
        CATCH(sample2 = [output copyNextSampleBuffer]);
        if (UNLIKELY(!sample2)) {
            DLOG("Couldn't retrieve video frame");
            goto error_free_output;
        }
        pts2 = CMSampleBufferGetPresentationTimeStamp(sample2);
        CFRelease(sample2);
    } while (CMTIME_COMPARE_INLINE(pts2, ==, pts1));
    if (UNLIKELY(CMTIME_COMPARE_INLINE(pts2, <, pts1))) {
        DLOG("First frame (%g) is later than second (%g), can't compute"
             " framerate", CMTimeGetSeconds(pts1), CMTimeGetSeconds(pts2));
        goto error_free_output;
    }
    const double frame_time = CMTimeGetSeconds(CMTimeSubtract(pts2, pts1));
    [output release];
    [reader release];

    /* Compute the framerate as the reciprocal of the frame time difference,
     * but attempt to detect common framerates and return them exactly. */
    const double computed_framerate = 1.0 / frame_time;
    double framerate;
    if (fabs(computed_framerate - round(computed_framerate)) < 1e-6) {
        framerate = round(computed_framerate);
        if (framerate != computed_framerate) {
            DLOG("Rounding framerate %.16g to %.0f",
                 computed_framerate, framerate);
        }
    } else if (fabs((computed_framerate * 1.001)
                    - round(computed_framerate * 1.001)) < 1e-6) {
        framerate = round(computed_framerate * 1.001) / 1.001;
        if (framerate != computed_framerate) {
            DLOG("Rounding framerate %.16g to %.0f/1.001",
                 computed_framerate, round(framerate * 1.001));
        }
    } else {
        framerate = computed_framerate;
    }

    return framerate;

  error_free_output:
    [output release];
  error_free_reader:
    [reader release];
  error_return:
    return 0;
}

/*-----------------------------------------------------------------------*/

static int movie_sound_open(SoundDecodeHandle *this)
{
    SysMovieHandle *movie = (SysMovieHandle *)this->custom_data;
    PRECOND(movie != NULL, return 0);

    this->reset       = movie_sound_reset;
    this->get_pcm     = movie_sound_get_pcm;
    this->close       = movie_sound_close;
    this->stereo      = (movie->audio_channels == 2);
    this->native_freq = movie->audio_rate;
    return 1;
}

/*-----------------------------------------------------------------------*/

static void movie_sound_reset(UNUSED SoundDecodeHandle *this)
{
    /* Nothing to do (should never be called). */
}

/*-----------------------------------------------------------------------*/

static int movie_sound_get_pcm(
    SoundDecodeHandle *this, int16_t *pcm_buffer, int pcm_len,
    int *loop_offset_ret)
{
    SysMovieHandle *movie = (SysMovieHandle *)this->custom_data;
    PRECOND(movie != NULL, return 0);

    *loop_offset_ret = 0;  // We don't loop.

    if (!movie->audio_reader) {
        return 0;
    }

    int pos = 0;
    while (pos < pcm_len) {
        if (movie->audio_offset < 0) {
            const int toclear = ubound(pcm_len - pos, -movie->audio_offset);
            mem_clear(&pcm_buffer[pos * movie->audio_channels],
                      toclear * (2*movie->audio_channels));
            pos += toclear;
            movie->audio_offset += toclear;
            if (pos >= pcm_len) {
                break;
            }
        }
        if (!movie->audio_buf) {
            if (movie->audio_reader.status != AVAssetReaderStatusReading) {
                break;
            }
            CMSampleBufferRef audio_buf = 0;
            CATCH(audio_buf = [movie->audio_output copyNextSampleBuffer]);
            if (!audio_buf) {
                break;
            }
            const CMBlockBufferRef block =
                CMSampleBufferGetDataBuffer(audio_buf);
            if (UNLIKELY(!block)) {
                DLOG("Failed to get audio block");
                CFRelease(audio_buf);
                break;
            }
            size_t length_at_offset = 0;  // Grab both lengths just to be safe.
            size_t length = 0;
            void *data = NULL;
            const OSStatus result = CMBlockBufferGetDataPointer(
                block, 0, &length_at_offset, &length, (char **)&data);
            if (UNLIKELY(result != kCMBlockBufferNoErr)) {
                DLOG("Failed to get audio data: %ld", (long)result);
                CFRelease(audio_buf);
                break;
            }
            ASSERT(length == length_at_offset, CFRelease(audio_buf); break);
            ASSERT(data != NULL, CFRelease(audio_buf); break);
            const int bytes_per_sample = movie->audio_channels * 2;
            ASSERT(length % bytes_per_sample == 0, CFRelease(audio_buf); break);
            movie->audio_buf = audio_buf;
            movie->audio_data = (int16_t *)data;
            movie->audio_buflen = length / bytes_per_sample;
            movie->audio_offset = 0;
        }
        const int tocopy =
            ubound(pcm_len - pos, movie->audio_buflen - movie->audio_offset);
        memcpy(&pcm_buffer[pos * movie->audio_channels],
               &movie->audio_data[movie->audio_offset * movie->audio_channels],
               tocopy * (2*movie->audio_channels));
        pos += tocopy;
        movie->audio_offset += tocopy;
        if (movie->audio_offset >= movie->audio_buflen) {
            CFRelease(movie->audio_buf);
            movie->audio_buf = NULL;
        }
    }

    return pos;
}

/*-----------------------------------------------------------------------*/

static void movie_sound_close(UNUSED SoundDecodeHandle *this)
{
    /* Nothing to do. */
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
