/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sound/decode.c: Tests for the audio decoding framework.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/memory.h"
#include "src/sound.h"
#include "src/sound/decode.h"
#include "src/sysdep.h"
#include "src/test/base.h"
#include "src/test/sound/wavegen.h"

#if defined(SIL_PLATFORM_ANDROID) || defined(SIL_PLATFORM_IOS) || defined(SIL_PLATFORM_LINUX) || defined(SIL_PLATFORM_MACOSX)
# define USING_IOQUEUE
# include "src/sysdep/posix/ioqueue.h"
#endif

/*************************************************************************/
/************************** Dummy decode module **************************/
/*************************************************************************/

/* Flag indicating whether decode_close() was called. */
static uint8_t decode_close_called;

/*-----------------------------------------------------------------------*/

static void decode_reset(SoundDecodeHandle *this)
{
    if (this->internal->data_type == SOUND_DECODE_CUSTOM) {
        SquareState *state = (SquareState *)this->custom_data;
        state->samples_out = 0;
    } else {
        this->private = (void *)0;
    }
}

/*-----------------------------------------------------------------------*/

static int decode_get_pcm(SoundDecodeHandle *this, int16_t *pcm_buffer,
                          int pcm_len, int *loop_offset_ret)
{
    *loop_offset_ret = 0;
    if (this->internal->data_type == SOUND_DECODE_CUSTOM) {
        SquareState *state = (SquareState *)this->custom_data;
        if (this->stereo) {
            return stereo_square_gen(state, pcm_buffer, pcm_len) ? pcm_len : 0;
        } else {
            return square_gen(state, pcm_buffer, pcm_len) ? pcm_len : 0;
        }
    } else {
        uintptr_t offset = (uintptr_t)this->private;
        const uint8_t *data;
        const uint32_t bytes_read =
            decode_get_data(this, offset, pcm_len, &data);
        for (uint32_t i = 0; i < bytes_read; i++) {
            if (this->stereo) {
                pcm_buffer[i*2+0] = data[i];
                pcm_buffer[i*2+1] = -pcm_buffer[i*2+0];
            } else {
                pcm_buffer[i] = data[i];
            }
        }
        offset += bytes_read;
        this->private = (void *)offset;
        return bytes_read;
    }
}

/*-----------------------------------------------------------------------*/

static void decode_close(UNUSED SoundDecodeHandle *this)
{
    decode_close_called = 1;
}

/*-----------------------------------------------------------------------*/

/* Magic numbers for looplen:
 *    42 = force failure (if using a memory or file source)
 *    47 = signal stereo rather than mono data
 *    111 = return frequency 0 (deliberately invalid value)
 */
static int decode_open(SoundDecodeHandle *this)
{
    if (this->internal->data_type == SOUND_DECODE_CUSTOM) {
        if (!this->custom_data) {
            return 0;
        }
    } else {
        if (this->internal->looplen == 42) {
            return 0;
        }
    }

    this->reset   = decode_reset;
    this->get_pcm = decode_get_pcm;
    this->close   = decode_close;
    this->stereo  = (this->internal->looplen == 47);
    this->native_freq = (this->internal->looplen == 111 ? 0 : 4);
    this->private = (void *)0;  // Sample counter for memory/file decodes.
    return 1;
}

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_sound_decode)

TEST_INIT(init)
{
    /* Override the default WAV handler so we have a format code we can
     * use for our tests. */
    sound_decode_set_handler(SOUND_FORMAT_WAV, decode_open);
    CHECK_TRUE(sound_decode_has_handler(SOUND_FORMAT_WAV));
    return 1;
}

TEST_CLEANUP(cleanup)
{
    /* Restore the original WAV decoder before returning. */
    sound_decode_set_handler(SOUND_FORMAT_WAV, decode_wav_open);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_handler)
{
    sound_decode_set_handler(SOUND_FORMAT_AUTODETECT, decode_open); // Invalid.
    CHECK_FALSE(sound_decode_has_handler(SOUND_FORMAT_AUTODETECT));
    sound_decode_set_handler(SOUND_FORMAT_WAV, NULL);
    CHECK_FALSE(sound_decode_has_handler(SOUND_FORMAT_WAV));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_decode_memory)
{
    static const uint8_t decode_data[] = {100, 100, 40, 40};

    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open(SOUND_FORMAT_WAV, decode_data,
                                           lenof(decode_data), 0, 0, 1));

    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);

    int16_t pcm[3];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);

    sound_decode_reset(decoder);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 2));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 1.0);
    CHECK_INTEQUAL(pcm[0], 40);
    CHECK_INTEQUAL(pcm[1], 0);
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 1.0);

    sound_decode_reset(decoder);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);

    decode_close_called = 0;
    sound_decode_close(decoder);
    CHECK_TRUE(decode_close_called);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_decode_memory_memory_failure)
{
    static const uint8_t decode_data[] = {100, 100, 40, 40};

    SoundDecodeHandle *decoder;
    CHECK_MEMORY_FAILURES(
        decoder = sound_decode_open(SOUND_FORMAT_WAV, decode_data,
                                    lenof(decode_data), 0, 0, 1));
    int16_t pcm[3];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);
    sound_decode_close(decoder);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_decode_file)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.dat"));

    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(SOUND_FORMAT_WAV, file,
                                                     1, 4, 0, 0, 1));
    sys_file_close(file);

    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);

    int16_t pcm[3];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);

    sound_decode_reset(decoder);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 2));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 1.0);
    CHECK_INTEQUAL(pcm[0], 40);
    CHECK_INTEQUAL(pcm[1], 0);
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 1.0);

    sound_decode_reset(decoder);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);

    decode_close_called = 0;
    sound_decode_close(decoder);
    CHECK_TRUE(decode_close_called);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_decode_file_memory_failure)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.dat"));

    SoundDecodeHandle *decoder;
    CHECK_MEMORY_FAILURES(
        decoder = sound_decode_open_from_file(SOUND_FORMAT_WAV, file,
                                              1, 4, 0, 0, 1));
    sys_file_close(file);
    int16_t pcm[3];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);
    sound_decode_close(decoder);

    return 1;
}

/*-----------------------------------------------------------------------*/

#ifdef USING_IOQUEUE
TEST(test_decode_file_read_permfail)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.dat"));

    /* Permanent failure should cause the open to fail. */
    TEST_posix_ioqueue_permfail_next_read(1);
    CHECK_FALSE(sound_decode_open_from_file(SOUND_FORMAT_WAV, file,
                                            1, 4, 0, 0, 1));

    sys_file_close(file);
    return 1;
}
#endif

/*-----------------------------------------------------------------------*/

#ifdef USING_IOQUEUE
TEST(test_decode_file_read_tempfail)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.dat"));
    SoundDecodeHandle *decoder;
    int16_t pcm[3];

    /* Transient failure should still allow the open to succeed. */
    TEST_posix_ioqueue_tempfail_next_read(1);
    CHECK_TRUE(decoder = sound_decode_open_from_file(SOUND_FORMAT_WAV, file,
                                                     1, 4, 0, 0, 1));
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);
    sound_decode_close(decoder);

    sys_file_close(file);
    return 1;
}
#endif

/*-----------------------------------------------------------------------*/

#ifdef USING_IOQUEUE
TEST(test_decode_file_read_tempfail_invalid_format)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.dat"));

    TEST_posix_ioqueue_tempfail_next_read(1);
    CHECK_FALSE(sound_decode_open_from_file(SOUND_FORMAT_AUTODETECT, file,
                                            1, 4, 0, 0, 1));

    sys_file_close(file);
    return 1;
}
#endif

/*-----------------------------------------------------------------------*/

#ifdef USING_IOQUEUE
TEST(test_decode_file_read_error)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.dat"));
    SoundDecodeHandle *decoder;
    int16_t pcm[3];

    /* I/O failure on the read should trigger an immediate read when
     * getting data. */
    TEST_posix_ioqueue_iofail_next_read(1);
    CHECK_TRUE(decoder = sound_decode_open_from_file(SOUND_FORMAT_WAV, file,
                                                     1, 4, 0, 0, 1));
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);
    sound_decode_close(decoder);

    sys_file_close(file);
    return 1;
}
#endif

/*-----------------------------------------------------------------------*/

TEST(test_decode_custom)
{
    SquareState state = {.period = 4, .num_cycles = 1, .samples_out = 0};
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_custom(decode_open, &state, 1));

    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);

    int16_t pcm[3];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 10000);
    CHECK_INTEQUAL(pcm[1], 10000);
    CHECK_INTEQUAL(pcm[2], -10000);

    sound_decode_reset(decoder);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 10000);
    CHECK_INTEQUAL(pcm[1], 10000);
    CHECK_INTEQUAL(pcm[2], -10000);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 2));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 1.25);
    CHECK_INTEQUAL(pcm[0], -10000);
    CHECK_INTEQUAL(pcm[1], 0);
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 1));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 1.25);

    sound_decode_reset(decoder);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 10000);
    CHECK_INTEQUAL(pcm[1], 10000);
    CHECK_INTEQUAL(pcm[2], -10000);

    decode_close_called = 0;
    sound_decode_close(decoder);
    CHECK_TRUE(decode_close_called);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_decode_custom_memory_failure)
{
    SquareState state = {.period = 4, .num_cycles = 1, .samples_out = 0};
    SoundDecodeHandle *decoder;
    CHECK_MEMORY_FAILURES(
        decoder = sound_decode_open_custom(decode_open, &state, 1));
    int16_t pcm[3];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_FLOATEQUAL(sound_decode_get_position(decoder), 0.75);
    CHECK_INTEQUAL(pcm[0], 10000);
    CHECK_INTEQUAL(pcm[1], 10000);
    CHECK_INTEQUAL(pcm[2], -10000);
    sound_decode_close(decoder);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resample)
{
    static const uint8_t decode_data[] = {100, 100, 40, 40};
    SoundDecodeHandle *decoder;
    int16_t pcm[8000];

    /* Test downsampling. */
    CHECK_TRUE(decoder = sound_decode_open(SOUND_FORMAT_WAV, decode_data,
                                           lenof(decode_data), 0, 0, 1));
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    sound_decode_set_output_freq(decoder, 2);
    /* This call should fail, leaving the output frequency at 2. */
    sound_decode_set_output_freq(decoder, 1);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 40);
    CHECK_INTEQUAL(pcm[2], 0);
    sound_decode_close(decoder);

    /* Test set_output_freq() to the native frequency. */
    CHECK_TRUE(decoder = sound_decode_open(SOUND_FORMAT_WAV, decode_data,
                                           lenof(decode_data), 0, 0, 1));
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    sound_decode_set_output_freq(decoder, 4);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 5));
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 40);
    CHECK_INTEQUAL(pcm[3], 40);
    CHECK_INTEQUAL(pcm[4], 0);
    sound_decode_close(decoder);

    /* Test upsampling. */
    CHECK_TRUE(decoder = sound_decode_open(SOUND_FORMAT_WAV, decode_data,
                                           lenof(decode_data), 0, 0, 1));
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    sound_decode_reset(decoder);
    sound_decode_set_output_freq(decoder, 5);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 6));
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 64);
    CHECK_INTEQUAL(pcm[3], 40);
    CHECK_INTEQUAL(pcm[4], 32);
    CHECK_INTEQUAL(pcm[5], 0);
    sound_decode_close(decoder);

    /* Test resampling in stereo. */
    CHECK_TRUE(decoder = sound_decode_open(SOUND_FORMAT_WAV, decode_data,
                                           lenof(decode_data), 0, 47, 1));
    CHECK_TRUE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    sound_decode_reset(decoder);
    sound_decode_set_output_freq(decoder, 5);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 6));
    CHECK_INTEQUAL(pcm[0*2+0], 100);
    CHECK_INTEQUAL(pcm[0*2+1], -100);
    CHECK_INTEQUAL(pcm[1*2+0], 100);
    CHECK_INTEQUAL(pcm[1*2+1], -100);
    CHECK_INTEQUAL(pcm[2*2+0], 64);
    CHECK_INTEQUAL(pcm[2*2+1], -64);
    CHECK_INTEQUAL(pcm[3*2+0], 40);
    CHECK_INTEQUAL(pcm[3*2+1], -40);
    CHECK_INTEQUAL(pcm[4*2+0], 32);
    CHECK_INTEQUAL(pcm[4*2+1], -32);
    CHECK_INTEQUAL(pcm[5*2+0], 0);
    CHECK_INTEQUAL(pcm[5*2+1], 0);
    sound_decode_close(decoder);

    /* Test handling of resample buffer reloads for large files. */
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/long.dat"));
    CHECK_TRUE(decoder = sound_decode_open_from_file(SOUND_FORMAT_WAV, file,
                                                     1, 39996, 0, 0, 1));
    sys_file_close(file);
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    sound_decode_set_output_freq(decoder, 1);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 8000));
    CHECK_INTEQUAL(pcm[0], 0);
    CHECK_INTEQUAL(pcm[7999], 15998%256);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 8000));
    CHECK_INTEQUAL(pcm[0], 16000%256);
    CHECK_INTEQUAL(pcm[1998], 19996%256);
    CHECK_INTEQUAL(pcm[1999], 0);
    CHECK_INTEQUAL(pcm[7999], 0);
    sound_decode_close(decoder);

    /* Test handling of a stream with no PCM data. */
    SquareState state = {.period = 4, .num_cycles = 0, .samples_out = 0};
    CHECK_TRUE(decoder = sound_decode_open_custom(decode_open, &state, 1));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    sound_decode_set_output_freq(decoder, 2);
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 3));
    sound_decode_close(decoder);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resample_no_interpolate)
{
    static const uint8_t decode_data[] = {100, 100, 40, 40};
    SoundDecodeHandle *decoder;
    int16_t pcm[8000];

    /* Test upsampling. */
    CHECK_TRUE(decoder = sound_decode_open(SOUND_FORMAT_WAV, decode_data,
                                           lenof(decode_data), 0, 0, 0));
    CHECK_FALSE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    sound_decode_reset(decoder);
    sound_decode_set_output_freq(decoder, 5);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 6));
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 100);
    CHECK_INTEQUAL(pcm[2], 100);
    CHECK_INTEQUAL(pcm[3], 40);
    CHECK_INTEQUAL(pcm[4], 40);
    CHECK_INTEQUAL(pcm[5], 0);
    sound_decode_close(decoder);

    /* Test resampling in stereo. */
    CHECK_TRUE(decoder = sound_decode_open(SOUND_FORMAT_WAV, decode_data,
                                           lenof(decode_data), 0, 47, 0));
    CHECK_TRUE(sound_decode_is_stereo(decoder));
    CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
    sound_decode_reset(decoder);
    sound_decode_set_output_freq(decoder, 5);
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 6));
    CHECK_INTEQUAL(pcm[0*2+0], 100);
    CHECK_INTEQUAL(pcm[0*2+1], -100);
    CHECK_INTEQUAL(pcm[1*2+0], 100);
    CHECK_INTEQUAL(pcm[1*2+1], -100);
    CHECK_INTEQUAL(pcm[2*2+0], 100);
    CHECK_INTEQUAL(pcm[2*2+1], -100);
    CHECK_INTEQUAL(pcm[3*2+0], 40);
    CHECK_INTEQUAL(pcm[3*2+1], -40);
    CHECK_INTEQUAL(pcm[4*2+0], 40);
    CHECK_INTEQUAL(pcm[4*2+1], -40);
    CHECK_INTEQUAL(pcm[5*2+0], 0);
    CHECK_INTEQUAL(pcm[5*2+1], 0);
    sound_decode_close(decoder);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resample_memory_failure)
{
    static const uint8_t decode_data[] = {100, 100, 40, 40};
    SoundDecodeHandle *decoder;
    int16_t pcm[3];

    /* This is a bit complicated for handling via CHECK_MEMORY_FAILURES(),
     * since set_output_freq() doesn't return a value and instead fakes
     * end-of-stream on memory allocation failure, so we check manually. */
    int i;
    for (i = 0; i < 100; i++) {
        CHECK_TRUE(decoder = sound_decode_open(SOUND_FORMAT_WAV, decode_data,
                                               lenof(decode_data), 0, 0, 1));
        CHECK_FALSE(sound_decode_is_stereo(decoder));
        CHECK_INTEQUAL(sound_decode_native_freq(decoder), 4);
        TEST_mem_fail_after(i, 1);
        sound_decode_set_output_freq(decoder, 2);
        TEST_mem_fail_after(-1, 0);
        if (sound_decode_get_pcm(decoder, pcm, 3)) {
            if (i == 0) {
                FAIL("sound_decode_set_output_freq(decoder, 2) did not fail"
                     " on a memory allocation failure");
            }
            break;
        }
        sound_decode_close(decoder);
    }
    if (i == 100) {
        FAIL("sound_decode_set_output_freq(decoder, 2) did not succeed"
             " after 100 iterations");
    }
    CHECK_INTEQUAL(pcm[0], 100);
    CHECK_INTEQUAL(pcm[1], 40);
    CHECK_INTEQUAL(pcm[2], 0);
    sound_decode_close(decoder);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_file_read)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/long.dat"));

    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(SOUND_FORMAT_WAV, file,
                                                     0, 40000, 0, 0, 1));
    sys_file_close(file);

    const uint8_t *data;

    /* Should be bounded to 16384 (READ_BUFFER_SIZE). */
    CHECK_INTEQUAL(decode_get_data(decoder, 0, 20000, &data), 16384);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 0);
    CHECK_INTEQUAL(data[16382]<<8 | data[16383], 8191);

    /* Completely separate part of the file. */
    CHECK_INTEQUAL(decode_get_data(decoder, 20000, 16384, &data), 16384);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 10000);
    CHECK_INTEQUAL(data[16382]<<8 | data[16383], 18191);

    /* Partially overlapping with the beginning of the buffered data. */
    CHECK_INTEQUAL(decode_get_data(decoder, 6000, 16384, &data), 16384);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 3000);
    CHECK_INTEQUAL(data[16382]<<8 | data[16383], 11191);

    /* Partially overlapping with the end of the buffered data. */
    CHECK_INTEQUAL(decode_get_data(decoder, 22000, 16384, &data), 16384);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 11000);
    CHECK_INTEQUAL(data[16382]<<8 | data[16383], 19191);

    /* Entirely within the end of the buffered data. */
    CHECK_INTEQUAL(decode_get_data(decoder, 36000, 2000, &data), 2000);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 18000);
    CHECK_INTEQUAL(data[1998]<<8 | data[1999], 18999);

    /* Small read (triggering read-ahead). */
    CHECK_INTEQUAL(decode_get_data(decoder, 0, 2000, &data), 2000);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 0);
    CHECK_INTEQUAL(data[1998]<<8 | data[1999], 999);

    /* Small read from read-ahead data. */
    CHECK_INTEQUAL(decode_get_data(decoder, 4000, 2000, &data), 2000);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 2000);
    CHECK_INTEQUAL(data[1998]<<8 | data[1999], 2999);

#ifdef USING_IOQUEUE
    /* Delayed I/O for read-ahead.  To trigger this, we block I/O and then
     * read from the end of the read buffer, triggering a read-ahead. */
    TEST_posix_ioqueue_block_io_thread(1);
    CHECK_INTEQUAL(decode_get_data(decoder, 10000, 2000, &data), 2000);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 5000);
    CHECK_INTEQUAL(data[1998]<<8 | data[1999], 5999);
    /* The read request will be detected as still pending here. */
    CHECK_INTEQUAL(decode_get_data(decoder, 12000, 2000, &data), 2000);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 6000);
    CHECK_INTEQUAL(data[1998]<<8 | data[1999], 6999);

    /* Small read from a different part of the file, cancelling the
     * pending read operation. */
    TEST_posix_ioqueue_unblock_on_wait(1);
    CHECK_INTEQUAL(decode_get_data(decoder, 36000, 2000, &data), 2000);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 18000);
    CHECK_INTEQUAL(data[1998]<<8 | data[1999], 18999);

    TEST_posix_ioqueue_unblock_on_wait(0);
    TEST_posix_ioqueue_block_io_thread(0);

    /* I/O error on immediate read. */
    TEST_posix_ioqueue_iofail_next_read(1);
    CHECK_FALSE(decode_get_data(decoder, 0, 2000, &data));
#endif

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_file_read_no_async_leak)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/long.dat"));

    /* Make sure that closing a decoder with an open async read request
     * doesn't leak the request.  We do this by filling up the async read
     * table except for 2 entries, operating on the decoder, and checking
     * that we can still create 2 new async requests after closing the
     * decoder. */

    char buf[1];
    int reqlist[1000];
    ASSERT(reqlist[0] = sys_file_read_async(file, buf, 1, 0));
    ASSERT(reqlist[1] = sys_file_read_async(file, buf, 1, 0));
    int i;
    for (i = 2; i < lenof(reqlist); i++) {
        if (!(reqlist[i] = sys_file_read_async(file, buf, 1, 0))) {
            break;
        }
    }
    if (i >= lenof(reqlist)) {
        FAIL("Unable to force sys_file_read_async() failure by running out"
             " of async read handles");
    }
    ASSERT(sys_file_wait_async(reqlist[--i]) == 1);
    ASSERT(sys_file_wait_async(reqlist[--i]) == 1);

    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_from_file(SOUND_FORMAT_WAV, file,
                                                     0, 40000, 0, 0, 1));

    /* Wait for the first async read to complete. */
    const uint8_t *data;
    CHECK_INTEQUAL(decode_get_data(decoder, 0, 2000, &data), 2000);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 0);
    CHECK_INTEQUAL(data[1998]<<8 | data[1999], 999);

    /* Trigger a read-ahead. */
    CHECK_INTEQUAL(decode_get_data(decoder, 10000, 2000, &data), 2000);
    CHECK_INTEQUAL(data[0]<<8 | data[1], 5000);
    CHECK_INTEQUAL(data[1998]<<8 | data[1999], 5999);

    sound_decode_close(decoder);
    CHECK_TRUE(reqlist[i] = sys_file_read_async(file, buf, 1, 0));
    CHECK_TRUE(reqlist[i+1] = sys_file_read_async(file, buf, 1, 0));
    CHECK_INTEQUAL(sys_file_wait_async(reqlist[i]), 1);
    CHECK_INTEQUAL(sys_file_wait_async(reqlist[i+1]), 1);

    for (i--; i >= 0; i--) {
        ASSERT(sys_file_wait_async(reqlist[i]) == 1);
    }

    sys_file_close(file);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_open_memory_invalid)
{
    static const uint8_t decode_data[] = {100};

    CHECK_FALSE(sound_decode_open(
                    SOUND_FORMAT_AUTODETECT, decode_data, 1, 0, 0, 1));
    CHECK_FALSE(sound_decode_open(SOUND_FORMAT_MP3, decode_data, 1, 0, 0, 1));
    CHECK_FALSE(sound_decode_open(SOUND_FORMAT_WAV, NULL, 1, 0, 0, 1));
    CHECK_FALSE(sound_decode_open(SOUND_FORMAT_WAV, decode_data, 0, 0, 0, 1));
    CHECK_FALSE(sound_decode_open(SOUND_FORMAT_WAV, decode_data, 1, 0, 42, 1));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_open_file_invalid)
{
    SysFile *file;
    ASSERT(file = sys_file_open("testdata/sound/square.dat"));

    CHECK_FALSE(sound_decode_open_from_file(
                    SOUND_FORMAT_AUTODETECT, file, 0, 1, 0, 0, 1));
    CHECK_FALSE(sound_decode_open_from_file(
                    SOUND_FORMAT_MP3, file, 0, 1, 0, 0, 1));
    CHECK_FALSE(sound_decode_open_from_file(
                    SOUND_FORMAT_WAV, NULL, 0, 1, 0, 0, 1));
    CHECK_FALSE(sound_decode_open_from_file(
                    SOUND_FORMAT_WAV, file, 0, 0, 0, 0, 1));
    CHECK_FALSE(sound_decode_open_from_file(
                    SOUND_FORMAT_WAV, file, 0, 1, 0, 42, 1));

    sys_file_close(file);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_open_custom_invalid)
{
    SquareState state = {.period = 4, .num_cycles = 1, .samples_out = 0};
    CHECK_FALSE(sound_decode_open_custom(NULL, &state, 1));
    CHECK_FALSE(sound_decode_open_custom(decode_open, NULL, 1));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_open_invalid_freq)
{
    static const uint8_t decode_data[] = {100};

    CHECK_FALSE(sound_decode_open(
                    SOUND_FORMAT_WAV, decode_data, 1, 0, 111, 1));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_output_freq_invalid)
{
    SquareState state = {.period = 4, .num_cycles = 1, .samples_out = 0};
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_custom(decode_open, &state, 1));

    sound_decode_set_output_freq(decoder, 0);

    int16_t pcm[3];
    CHECK_TRUE(sound_decode_get_pcm(decoder, pcm, 3));
    CHECK_INTEQUAL(pcm[0], 10000);
    CHECK_INTEQUAL(pcm[1], 10000);
    CHECK_INTEQUAL(pcm[2], -10000);

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_get_pcm_invalid)
{
    SquareState state = {.period = 4, .num_cycles = 1, .samples_out = 0};
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_custom(decode_open, &state, 1));

    int16_t pcm[1];
    CHECK_FALSE(sound_decode_get_pcm(decoder, NULL, 1));
    CHECK_FALSE(sound_decode_get_pcm(decoder, pcm, 0));

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_get_data_bad_type)
{
    SquareState state = {.period = 4, .num_cycles = 1, .samples_out = 0};
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_custom(decode_open, &state, 1));

    /* Muck with the data_type field to cover the default code path for
     * the switch (this->internal->data_type) in decode_get_data().  (This
     * path should never be hit outside testing; we run it only to improve
     * coverage.) */
    ASSERT(decoder->internal->data_type == SOUND_DECODE_CUSTOM);
    decoder->internal->data_type = SOUND_DECODE_BUFFER - 1;  // Invalid valid.
    decoder->internal->datalen = 1;
    const uint8_t *data;
    CHECK_INTEQUAL(decode_get_data(decoder, 0, 1, &data), 1);
    decoder->internal->datalen = 0;
    decoder->internal->data_type = SOUND_DECODE_CUSTOM;

    sound_decode_close(decoder);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_get_data_invalid)
{
    SquareState state = {.period = 4, .num_cycles = 1, .samples_out = 0};
    SoundDecodeHandle *decoder;
    CHECK_TRUE(decoder = sound_decode_open_custom(decode_open, &state, 1));

    const uint8_t *data;
    CHECK_FALSE(decode_get_data(decoder, 0, 1, &data));

    decoder->internal->datalen = 1;
    CHECK_FALSE(decode_get_data(decoder, 0, 1, &data));
    decoder->internal->datalen = 0;

    sound_decode_close(decoder);
    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
