/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/posix/internal.c: Helper functions for POSIX
 * file-related tests.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/random.h"
#include "src/test/base.h"
#include "src/test/sysdep/posix/internal.h"

#ifdef SIL_PLATFORM_ANDROID
# include "src/sysdep.h"
# include "src/sysdep/android/internal.h"
#endif

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>

/*************************************************************************/
/*************************************************************************/

const char *posix_get_tmpdir(void)
{
#ifdef SIL_PLATFORM_ANDROID
    /* $TMPDIR doesn't seem to be defined on Android, and /tmp doesn't
     * exist either, so fall back to the internal data directory. */
    ASSERT(android_internal_data_path != NULL);
    return android_internal_data_path;
#else
    const char *tmpdir = getenv("TMPDIR");
    if (tmpdir && *tmpdir) {
        return tmpdir;
    } else {
        return "/tmp";
    }
#endif
}

/*-----------------------------------------------------------------------*/

int posix_create_temporary_dir(const char *basename, char *pathbuf,
                               int pathbuf_size)
{
    PRECOND(basename != NULL);
    PRECOND(pathbuf != NULL);
    PRECOND(pathbuf_size > 0);

    srandom_env();
    const char *tmpdir = posix_get_tmpdir();
    for (int tries = 10; tries > 0; tries--) {
        if (!strformat_check(pathbuf, pathbuf_size - 1,
                             "%s%s%s_%u_%d",
                             tmpdir, tmpdir[strlen(tmpdir)-1]=='/' ? "" : "/",
                             basename, getpid(), random())) {
            DLOG("Buffer overflow generating temporary path (tmpdir=%s)",
                 tmpdir);
            return 0;
        }
        if (mkdir(pathbuf, S_IRWXU) == 0) {
            return 1;
        }
        DLOG("Failed to create temporary directory %s: %s", pathbuf,
             strerror(errno));
    }
    DLOG("Unable to create temporary pipe, giving up");
    return 0;
}

/*-----------------------------------------------------------------------*/

int posix_pipe_writer(void *path)
{
    void (*old_sigpipe)(int) = signal(SIGPIPE, SIG_IGN);
    nanosleep(&(struct timespec){.tv_sec = 0, .tv_nsec = 10*1000*1000}, NULL);
    const int wfd = open((const char *)path, O_WRONLY);
    ASSERT(wfd >= 0);
    int bytes_written = write(wfd, "foo", 3);
    CHECK_INTEQUAL(close(wfd), 0);
    signal(SIGPIPE, old_sigpipe);
    return bytes_written > 0 ? bytes_written : 0;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
